import "reflect-metadata";
import { DeserializerContext } from "./deserializer";
import { SerializerContext } from "./serializer";
declare type FactoryReader = (value: string, ctx: DeserializerContext) => any;
declare type FactoryWriter = (value: any, ctx: SerializerContext) => string;
declare type FactoryTuple = [FactoryReader, FactoryWriter];
declare type TypeGetter = () => Function;
declare type IsTypeGetter = (o: any) => boolean;
export interface RootOptions {
    /**
     * Unqualified root element name.
     */
    name?: string;
    /**
     * Root element namespace URI.
     */
    namespaceUri?: string;
}
export interface ElementOptions {
    types?: ArrayItemOptions[];
}
export interface ArrayOptions {
    /**
     * Unqualified container element name. Only used when nested is true.
     */
    name?: string;
    /**
     * Container element namespace URI. Only used when nested is true.
     */
    namespaceUri?: string;
    /**
     * Indicates whether the array is wrapped in a container XML element.
     */
    nested?: boolean;
    /**
     * Array of possible element types this array can contain.
     */
    itemTypes?: ArrayItemOptions[];
}
export interface ArrayItemOptions {
    /**
     * Unqualified element name.
     */
    name?: string;
    /**
     * Element namespace URI.
     */
    namespaceUri?: string;
    /**
     * A method which returns the array item type.
     * (The item type cannot be derived from the decorator metadata. The item type is a callback to support cyclic references to types defined later in the file)
     */
    itemType?: TypeGetter;
    /**
     * A method which returns whether the JavaScript object resolves to the item type.
     */
    isType?: IsTypeGetter;
}
export interface AttributeOptions {
    /**
     * Unqualified attribute name.
     */
    name?: string;
    /**
     * Attribute namespace URI.
     */
    namespaceUri?: string;
    /**
     * Optional factory tuple to deserialize and serialize the attribute value.
     */
    factory?: FactoryTuple;
    /**
     * Optional attribute type.
     */
    type?: Function;
}
export interface TextOptions {
}
export interface BaseSchema {
    xmlType: "root" | "element" | "attribute" | "array" | "text";
}
export interface RootSchema extends BaseSchema {
    xmlType: "root";
    /**
     * The unqualified XML node name. Must not specify a prefix.
     */
    name: string;
    /**
     * The XML namespace URI for this node. Default for elements: "". Empty string.
     */
    namespaceUri: string | null;
    type: any;
}
export interface ElementSchema extends BaseSchema {
    xmlType: "element";
    propertyKey: string;
    types: ArrayItemOptions[];
}
export interface TextSchema extends BaseSchema {
    xmlType: "text";
    propertyKey: string;
    type: Function;
}
export interface AttributeSchema extends BaseSchema {
    xmlType: "attribute";
    propertyKey: string;
    /**
     * The unqualified XML node name. Must not specify a prefix.
     */
    name: string;
    /**
     * The XML namespace URI for this node.
     * Default for attributes: null. Implies it belongs to the element.
     */
    namespaceUri: string | null;
    factory?: FactoryTuple;
    type: Function;
}
export interface ArraySchema extends BaseSchema {
    xmlType: "array";
    propertyKey: string;
    /**
     * The unqualified XML node name. Must not specify a prefix.
     */
    name: string;
    /**
     * The XML namespace URI for this node. Default for elements: "". Empty string.
     */
    namespaceUri: string | null;
    /**
     * Indicates whether the array is wrapped in a container XML element.
     */
    nested: boolean;
    itemTypes: ArrayItemOptions[];
}
export declare function XMLRoot(opts?: RootOptions): (target: Function) => void;
export declare function XMLElement<T>(opts?: ElementOptions): (target: T, propertyKey: string) => void;
export declare function XMLAttribute(opts?: AttributeOptions): (target: any, propertyKey: string) => void;
export declare function XMLArray(opts?: ArrayOptions): (target: any, propertyKey: string) => void;
export declare function isArraySchema(schema: BaseSchema): schema is ArraySchema;
export declare function isRootSchema(schema: BaseSchema): schema is RootSchema;
export declare function isElementSchema(schema: BaseSchema): schema is ElementSchema;
export declare function isAttributeSchema(schema: BaseSchema): schema is AttributeSchema;
export declare function isTextSchema(schema: BaseSchema): schema is TextSchema;
export declare function XMLText(opts?: TextOptions): (target: any, propertyKey: string) => void;
export {};
