"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
function XMLRoot(opts = {}) {
    return function (target) {
        const rootSchema = {
            name: opts.name || target.name,
            namespaceUri: opts.namespaceUri || "",
            xmlType: "root",
            type: target,
        };
        let targetSchema = Reflect.getMetadata("xml:root", target) || [];
        if (targetSchema.length === 0) {
            Reflect.defineMetadata("xml:root", targetSchema, target);
        }
        targetSchema.push(rootSchema);
    };
}
exports.XMLRoot = XMLRoot;
function XMLElement(opts = {}) {
    return function (target, propertyKey) {
        let type = Reflect.getMetadata("design:type", target, propertyKey);
        if (type === Array) {
            throw new Error("@XMLElement decorator does not support array types on " + propertyKey + " in " + target.constructor.name);
        }
        let targetChildren = Reflect.getMetadata("xml:type:children", target.constructor) || [];
        if (targetChildren.length === 0) {
            Reflect.defineMetadata("xml:type:children", targetChildren, target.constructor);
        }
        targetChildren.push({
            propertyKey: propertyKey,
            xmlType: "element",
            types: !!opts.types ? getElementTypes(propertyKey, opts.types, type) : [{ itemType: () => type, name: propertyKey, namespaceUri: "" }],
        });
    };
}
exports.XMLElement = XMLElement;
function XMLAttribute(opts = {}) {
    return function (target, propertyKey) {
        let type = Reflect.getMetadata("design:type", target, propertyKey);
        if ((!opts.type && !opts.factory) && type === Object) {
            throw new Error("@XMLAttribute must specify type or factory on " + propertyKey + " in " + target.constructor.name);
        }
        let targetChildren = Reflect.getMetadata("xml:type:children", target.constructor) || [];
        if (targetChildren.length === 0) {
            Reflect.defineMetadata("xml:type:children", targetChildren, target.constructor);
        }
        targetChildren.push({
            propertyKey: propertyKey,
            factory: opts.factory,
            name: opts.name || propertyKey || "",
            namespaceUri: opts.namespaceUri || null,
            type: opts.type || type,
            xmlType: "attribute",
        });
    };
}
exports.XMLAttribute = XMLAttribute;
function XMLArray(opts = {}) {
    return function (target, propertyKey) {
        const type = Reflect.getMetadata("design:type", target, propertyKey);
        if (type !== Array) {
            throw new Error("@XMLArray requires an array type on " + propertyKey + " in " + target.constructor.name);
        }
        const targetChildren = Reflect.getMetadata("xml:type:children", target.constructor) || [];
        if (targetChildren.length === 0) {
            Reflect.defineMetadata("xml:type:children", targetChildren, target.constructor);
        }
        const name = opts.name || propertyKey;
        const nested = opts.nested !== undefined ? opts.nested : true;
        const arraySchema = {
            propertyKey: propertyKey,
            name: name,
            namespaceUri: opts.namespaceUri || "",
            xmlType: "array",
            nested: nested,
            itemTypes: !!opts.itemTypes ? getArrayItemTypes(propertyKey, opts.itemTypes, type) : [],
        };
        targetChildren.push(arraySchema);
    };
}
exports.XMLArray = XMLArray;
function isArraySchema(schema) {
    return schema.xmlType == "array";
}
exports.isArraySchema = isArraySchema;
function isRootSchema(schema) {
    return schema.xmlType == "root";
}
exports.isRootSchema = isRootSchema;
function isElementSchema(schema) {
    return schema.xmlType == "element";
}
exports.isElementSchema = isElementSchema;
function isAttributeSchema(schema) {
    return schema.xmlType == "attribute";
}
exports.isAttributeSchema = isAttributeSchema;
function isTextSchema(schema) {
    return schema.xmlType == "text";
}
exports.isTextSchema = isTextSchema;
function XMLText(opts = {}) {
    return function (target, propertyKey) {
        const type = Reflect.getMetadata("design:type", target, propertyKey);
        const targetChildren = Reflect.getMetadata("xml:type:children", target.constructor) || [];
        if (targetChildren.length === 0) {
            Reflect.defineMetadata("xml:type:children", targetChildren, target.constructor);
        }
        const textSchema = {
            propertyKey: propertyKey,
            type: type,
            xmlType: "text",
        };
        targetChildren.push(textSchema);
    };
}
exports.XMLText = XMLText;
function getElementTypes(propertyKey, types, fallbackType) {
    // Throw if types have duplicate names; the name is the xml disambiguator
    // TODO: check default names too (later)
    for (let type of types) {
        if (type.name && types.find(t => t !== type && t.name === type.name)) {
            throw new Error("@XMLElement cannot declare duplicate element name " + type.name + " on " + propertyKey);
        }
    }
    // If there is a single type without itemType, use the default fallback type from decorator metadata
    return types.map(t => ({
        name: t.name || null,
        namespaceUri: t.namespaceUri || "",
        isType: t.isType,
        itemType: t.itemType || (types.length === 1 ? () => fallbackType : undefined),
    }));
}
function getArrayItemTypes(propertyKey, types, fallbackType) {
    // Throw if types have duplicate names; the name is the xml disambiguator
    // TODO: check default names too (later)
    for (let type of types) {
        if (type.name && types.find(t => t !== type && t.name === type.name)) {
            throw new Error("@XMLArray cannot declare duplicate element name " + type.name + " on " + propertyKey);
        }
        if (!type.itemType) {
            throw new Error("@XMLArray must declare item type on " + propertyKey);
        }
    }
    return types.map(t => ({
        name: t.name || null,
        namespaceUri: t.namespaceUri || "",
        isType: t.isType,
        itemType: t.itemType,
    }));
}
//# sourceMappingURL=decorators.js.map