"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const decorators_1 = require("./decorators");
const xmldom_1 = require("xmldom");
const deserializer_1 = require("./deserializer");
class XMLDecoratorSerializer {
    constructor() {
        this.factory = new xmldom_1.DOMImplementation();
        this.document = null;
        this.prefixCounter = 0;
        this.prefixMap = {};
    }
    serialize(data, type, defaultNSPrefixMap, elementName, namespaceUri) {
        this.prefixCounter = 0;
        this.prefixMap = Object.assign({}, (defaultNSPrefixMap || { "": "" }));
        const xmlRootSchemas = Reflect.getMetadata("xml:root", type);
        if (!xmlRootSchemas) {
            throw new Error("Root type must specify @XMLRoot decorator");
        }
        if (xmlRootSchemas.length > 1 && !elementName) {
            throw new Error("Element name parameter is required when root type specifies multiple @XMLRoot decorators.");
        }
        let xmlSchema = null;
        if (xmlRootSchemas.length == 1 && !elementName) {
            xmlSchema = xmlRootSchemas[0];
        }
        else {
            for (let schema of xmlRootSchemas) {
                if (elementName === schema.name && (namespaceUri ? namespaceUri : "") == schema.namespaceUri) {
                    xmlSchema = schema;
                    break;
                }
            }
        }
        if (!xmlSchema) {
            throw new Error("Root type does not specify a @XMLRoot decorator for " + elementName);
        }
        // const xmlSchema = xmlRootSchemas[0];
        this.document = this.factory.createDocument(xmlSchema.namespaceUri, this.getQualifiedName(xmlSchema.name, xmlSchema.namespaceUri || ""), null);
        this.serializeObject(null, type, data, xmlSchema.name, xmlSchema.namespaceUri || "");
        if (!this.document) {
            throw new Error("Internal error. Document is null.");
        }
        const serializer = new xmldom_1.XMLSerializer();
        const result = serializer.serializeToString(this.document);
        this.document = null;
        return result;
    }
    getQualifiedAttributeName(attributeName, namespaceUri) {
        // null namespaceUri = scoped to element, no prefix
        if (namespaceUri === null) {
            return attributeName;
        }
        return this.getQualifiedName(attributeName, namespaceUri);
    }
    getQualifiedName(elementName, namespaceUri) {
        if (!this.prefixMap.hasOwnProperty(namespaceUri)) {
            this.prefixMap[namespaceUri] = "p" + this.prefixCounter;
            this.prefixCounter++;
        }
        var prefix = this.prefixMap[namespaceUri];
        return prefix.length > 0 ? (prefix + ":" + elementName) : elementName;
    }
    serializeObject(parentNode, type, data, elementName, namespaceUri) {
        if (!this.document) {
            throw new Error("Internal error. Document is null.");
        }
        let element;
        if (parentNode === null) {
            element = this.document.documentElement;
        }
        else {
            element = this.document.createElementNS(namespaceUri, this.getQualifiedName(elementName, namespaceUri));
            parentNode.appendChild(element);
        }
        const children = Reflect.getMetadata("xml:type:children", type) || [];
        for (let child of children) {
            if (decorators_1.isAttributeSchema(child)) {
                let value = data[child.propertyKey];
                if (value !== undefined) {
                    if (child.factory) {
                        value = child.factory[1](value, this);
                    }
                    else {
                        value = this.convertValue(value, child.type);
                    }
                }
                if (value !== undefined) {
                    const attrName = this.getQualifiedAttributeName(child.name, child.namespaceUri);
                    element.setAttributeNS(child.namespaceUri, attrName, value);
                }
            }
            else if (decorators_1.isTextSchema(child)) {
                const value = this.convertValue(data[child.propertyKey], child.type);
                if (value !== undefined) {
                    element.appendChild(this.document.createTextNode(value));
                }
            }
        }
        for (let child of children) {
            if (decorators_1.isElementSchema(child)) {
                const value = data[child.propertyKey];
                if (value === undefined) {
                    // TODO: Throw if not optional
                    continue;
                }
                const elementItemType = this.getArrayItemType(child.types, value);
                if (!elementItemType || !elementItemType.itemType) {
                    throw new Error("Internal error " + JSON.stringify(child));
                }
                const elementType = elementItemType.itemType();
                this.serializeElement(element, elementItemType.name || elementType.name, elementItemType.namespaceUri || "", elementType, value);
            }
            else if (decorators_1.isArraySchema(child)) {
                const value = data[child.propertyKey];
                if (value === undefined) {
                    // TODO: Throw if not optional
                    continue;
                }
                this.serializeArray(element, child, value);
            }
        }
    }
    convertValue(value, type) {
        if (value === undefined) {
            return;
        }
        if (type === String) {
            return value;
        }
        else if (type === Number) {
            return value.toString();
        }
        else if (type === Boolean) {
            return value.toString();
        }
        else if (type === Date) {
            return value.toISOString();
        }
        else {
            throw new Error("Cannot convert value " + value + " of " + type);
        }
    }
    serializeElement(parentNode, name, namespaceUri, elementType, data) {
        if (!this.document || !elementType) {
            throw new Error("Internal error.");
        }
        if (elementType === Number || elementType === Boolean || elementType === String || elementType === Date) {
            this.serializeValueElement(parentNode, name, namespaceUri, elementType, data);
        }
        else if (typeof elementType === "function") {
            this.serializeObject(parentNode, elementType, data, name, namespaceUri);
        }
        else {
            throw new Error("Cannot serialize type " + elementType);
        }
    }
    serializeValueElement(parentNode, name, namespaceUri, elementType, data) {
        if (!this.document) {
            throw new Error("Internal error.");
        }
        const element = this.document.createElementNS(namespaceUri, this.getQualifiedName(name, namespaceUri));
        element.appendChild(this.document.createTextNode(this.convertValue(data, elementType)));
        parentNode.appendChild(element);
    }
    getArrayItemType(itemTypes, dataItem) {
        // First check if the data constructor matches any of the item types
        for (let schemaItemType of itemTypes) {
            if (!schemaItemType.itemType) {
                throw new Error("itemType must be specified. " + +JSON.stringify(schemaItemType));
            }
            if (dataItem.constructor === schemaItemType.itemType()) {
                return schemaItemType;
            }
        }
        for (let schemaItemType of itemTypes) {
            if (!schemaItemType.itemType) {
                throw new Error("itemType must be specified. " + +JSON.stringify(schemaItemType));
            }
            if (itemTypes.length === 1) {
                return schemaItemType;
            }
            if (!schemaItemType.isType) {
                throw new Error("When serializing a value with multiple types, the value must have a matching constructor or specify a callback for isType. " + JSON.stringify(schemaItemType));
            }
            if (schemaItemType.isType(dataItem)) {
                return schemaItemType;
            }
        }
        return null;
    }
    serializeArray(parentNode, schema, data) {
        if (!this.document || schema.namespaceUri === null) {
            throw new Error("Internal error.");
        }
        if (schema.nested) {
            var nestedNode = this.document.createElementNS(schema.namespaceUri, this.getQualifiedName(schema.name, schema.namespaceUri));
            parentNode.appendChild(nestedNode);
            parentNode = nestedNode;
        }
        for (var i = 0; i < data.length; i++) {
            const dataItem = data[i];
            const dataItemType = this.getArrayItemType(schema.itemTypes, dataItem);
            if (!dataItemType) {
                throw new Error("Cannot find type for array item index " + i);
            }
            if (!dataItemType.itemType) {
                throw new Error("Cannot find itemType for array item index " + i);
            }
            const itemName = deserializer_1.getArrayItemName(schema, dataItemType);
            this.serializeElement(parentNode, dataItemType.name || itemName, dataItemType.namespaceUri || "", dataItemType.itemType(), data[i]);
        }
    }
}
exports.XMLDecoratorSerializer = XMLDecoratorSerializer;
//# sourceMappingURL=serializer.js.map